@@ -1,41 +1,136 @@
+import os
+from collections.abc import Generator
+from typing import Callable, Optional
+
 import numpy as np
 import pandas
-import os
 
 try:
     from os import scandir
 except ImportError:
-    from scandir import scandir
+    from scandir import scandir  # type: ignore
 
 
-class PyFileIndex(object):
+class PyFileIndex:
     """
-    The PyFileIndex maintains a pandas Dataframe to track changes in the file system.
+    The PyFileIndex maintains a pandas DataFrame to track changes in the file system.
 
     Args:
         path (str): file system path
-        filter_function (function): function to filter for specific files
-        debug (bool): enable debug print statements
+        filter_function (Callable): function to filter for specific files (optional)
+        debug (bool): enable debug print statements (optional)
+        df (pandas.DataFrame): DataFrame of a previous PyFileIndex object (optional)
     """
 
-    def __init__(self, path=".", filter_function=None, debug=False):
+    def __init__(
+        self,
+        path: str = ".",
+        filter_function: Optional[Callable] = None,
+        debug: bool = False,
+        df: Optional[pandas.DataFrame] = None,
+    ) -> None:
+        abs_path = os.path.abspath(os.path.expanduser(path))
+        self._check_if_path_exists(path=abs_path)
         self._debug = debug
         self._filter_function = filter_function
-        path = os.path.abspath(path)
-        self._df = self._create_df_from_lst(
-            list([self._get_lst_entry_from_path(entry=path)])
-            + list(self._scandir(path=path, df=None, recursive=True))
-        )
+        self._path = abs_path
+        if df is None:
+            self._df = self._create_df_from_lst(
+                [self._get_lst_entry_from_path(entry=self._path)]
+                + list(self._scandir(path=self._path, df=None, recursive=True))
+            )
+        else:
+            self._df = df
 
     @property
-    def df(self):
+    def df(self) -> pandas.DataFrame:
         return self._df
 
     @property
-    def dataframe(self):
+    def dataframe(self) -> pandas.DataFrame:
         return self.df
 
-    def _init_df_lst(self, path_lst, df=None, include_root=True):
+    def open(self, path: str) -> "PyFileIndex":
+        """
+        Open PyFileIndex in the subdirectory path
+
+        Args:
+             path (str): subdirectory to open
+
+        Returns:
+            PyFileIndex: PyFileIndex for subdirectory
+        """
+        abs_path = os.path.abspath(os.path.expanduser(os.path.join(self._path, path)))
+        self._check_if_path_exists(path=abs_path)
+        self.update()
+        if abs_path == self._path:
+            return self
+        elif (
+            os.path.commonpath([abs_path, self._path]) == self._path and os.name != "nt"
+        ):
+            return PyFileIndex(
+                path=abs_path,
+                filter_function=self._filter_function,
+                debug=self._debug,
+                df=self._df[self._df["path"].str.contains(abs_path)],
+            )
+        elif (
+            os.path.commonpath([abs_path, self._path]) == self._path and os.name != "nt"
+        ):
+            abs_path_unix = abs_path.replace("\\", "/")
+            return PyFileIndex(
+                path=abs_path,
+                filter_function=self._filter_function,
+                debug=self._debug,
+                df=self._df[
+                    np.array(
+                        [
+                            p.replace("\\", "/").contains(abs_path_unix)
+                            for p in self._df["path"].values
+                        ]
+                    )
+                ],
+            )
+        else:
+            return PyFileIndex(
+                path=abs_path,
+                filter_function=self._filter_function,
+                debug=self._debug,
+            )
+
+    def update(self) -> None:
+        """
+        Update file index
+        """
+        self._check_if_path_exists(path=self._path)
+        df_new, files_changed_lst, path_deleted_lst = self._get_changes_quick()
+        if self._debug:
+            print("Changes: ", df_new.path.values, files_changed_lst, path_deleted_lst)
+        if len(path_deleted_lst) != 0:
+            self._df = self._df[~self._df.path.isin(path_deleted_lst)]
+        if len(files_changed_lst) != 0:
+            df_updated = self._create_df_from_lst(
+                [self._get_lst_entry_from_path(entry=f) for f in files_changed_lst]
+            )
+            self._df = self._df[~self._df.path.isin(df_updated.path)]
+            self._df = (
+                pandas.concat([self._df, df_updated])
+                .drop_duplicates()
+                .reset_index(drop=True)
+            )
+        if len(df_new) != 0:
+            self._df = (
+                pandas.concat([self._df, df_new])
+                .drop_duplicates()
+                .reset_index(drop=True)
+            )
+
+    def _init_df_lst(
+        self,
+        path_lst: list,
+        df: Optional[pandas.DataFrame] = None,
+        include_root: bool = True,
+    ) -> pandas.DataFrame:
         """
         Internal function to build the pandas file index from a list of directories
 
@@ -55,14 +150,16 @@ class PyFileIndex(object):
                 total_lst.append(entry)
         return self._create_df_from_lst(total_lst)
 
-    def _scandir(self, path, df=None, recursive=True):
+    def _scandir(
+        self, path: str, df: Optional[pandas.DataFrame] = None, recursive: bool = True
+    ) -> Generator:
         """
-        Internal function to recursivley scan directories
+        Internal function to recursively scan directories
 
         Args:
             path (str): file system path
             df (pandas.DataFrame/ None): existing file index table
-            recursive (bool): recusivley iterate over sub directories
+            recursive (bool): recursively iterate over subdirectories
 
         Returns:
             list: list of file entries
@@ -86,14 +183,14 @@ class PyFileIndex(object):
                     else:
                         yield self._get_lst_entry(entry=entry)
         except FileNotFoundError:
-            return []
+            yield from ()
 
-    def _get_changes_quick(self):
+    def _get_changes_quick(self) -> tuple:
         """
         Internal function to list the changes to the file system
 
         Returns:
-            list: pandas.DataFrame with new entries, list of changed files and list of deleted paths
+            tuple: pandas.DataFrame with new entries, list of changed files, and list of deleted paths
         """
         path_exists_bool_lst = [os.path.exists(p) for p in self._df.path.values]
         path_deleted_lst = self._df[~np.array(path_exists_bool_lst)].path.values
@@ -118,33 +215,7 @@ class PyFileIndex(object):
         )
         return df_new, files_changed_lst, path_deleted_lst
 
-    def update(self):
-        """
-        Update file index
-        """
-        df_new, files_changed_lst, path_deleted_lst = self._get_changes_quick()
-        if self._debug:
-            print("Changes: ", df_new.path.values, files_changed_lst, path_deleted_lst)
-        if len(path_deleted_lst) != 0:
-            self._df = self._df[~self._df.path.isin(path_deleted_lst)]
-        if len(files_changed_lst) != 0:
-            df_updated = self._create_df_from_lst(
-                [self._get_lst_entry_from_path(entry=f) for f in files_changed_lst]
-            )
-            self._df = self._df[~self._df.path.isin(df_updated.path)]
-            self._df = (
-                pandas.concat([self._df, df_updated])
-                .drop_duplicates()
-                .reset_index(drop=True)
-            )
-        if len(df_new) != 0:
-            self._df = (
-                pandas.concat([self._df, df_new])
-                .drop_duplicates()
-                .reset_index(drop=True)
-            )
-
-    def _get_lst_entry_from_path(self, entry):
+    def _get_lst_entry_from_path(self, entry: str) -> list:
         """
         Internal function to generate file index entry from file system path
 
@@ -175,7 +246,7 @@ class PyFileIndex(object):
         except FileNotFoundError:
             return []
 
-    def _get_lst_entry(self, entry):
+    def _get_lst_entry(self, entry) -> list:
         """
         Internal function to generate file index entry from scandir DirEntry
 
@@ -206,17 +277,33 @@ class PyFileIndex(object):
         except FileNotFoundError:
             return []
 
-    def _repr_html_(self):
+    def _repr_html_(self) -> str:
         """
-        Internal visualisation function for iPython notebooks
+        Internal visualization function for iPython notebooks
 
         Returns:
-            HTML object: iPython notebook representation of the pandas.DataFrame
+            str: iPython notebook representation of the pandas.DataFrame
         """
         return self._df._repr_html_()
 
     @staticmethod
-    def _create_df_from_lst(lst):
+    def _check_if_path_exists(path: str) -> None:
+        """
+        Internal function to check if the given path exists
+
+        Args:
+            path (str): file system path
+
+        Raises:
+            FileNotFoundError: if the path does not exist
+        """
+        if not os.path.exists(path):
+            raise FileNotFoundError(
+                "The path " + path + " does not exist on your filesystem."
+            )
+
+    @staticmethod
+    def _create_df_from_lst(lst: list) -> pandas.DataFrame:
         """
         Internal function to generate file index as pandas from a list of entries
 
@@ -226,19 +313,19 @@ class PyFileIndex(object):
         Returns:
             pandas.DataFrame: file index
         """
-        lst_clean = [l for l in lst if len(l) != 0]
+        lst_clean = [sub_lst for sub_lst in lst if len(sub_lst) != 0]
         if len(lst_clean) != 0:
             name_lst, path_lst, dirname_lst, dir_lst, mtime_lst, nlink_lst = zip(
                 *lst_clean
             )
         else:
             name_lst, path_lst, dirname_lst, dir_lst, mtime_lst, nlink_lst = (
-                [],
-                [],
-                [],
-                [],
-                [],
-                [],
+                (),
+                (),
+                (),
+                (),
+                (),
+                (),
             )
         return pandas.DataFrame(
             {
@@ -250,3 +337,12 @@ class PyFileIndex(object):
                 "nlink": nlink_lst,
             }
         )
+
+    def __len__(self) -> int:
+        """
+        Returns the number of files in the index.
+
+        Returns:
+            int: The number of files in the index.
+        """
+        return len(self._df[~self._df.is_directory])
